import os
import sys
from ctypes import *
from ctypes.util import find_library

import distutils.sysconfig





def get_shared_lib_extension(is_python_ext=False):
    """Return the correct file extension for shared libraries.

    Parameters
    ----------
    is_python_ext : bool, optional
        Whether the shared library is a Python extension.  Default is False.

    Returns
    -------
    so_ext : str
        The shared library extension.

    Notes
    -----
    For Python shared libs, `so_ext` will typically be '.so' on Linux and OS X,
    and '.pyd' on Windows.  For Python >= 3.2 `so_ext` has a tag prepended on
    POSIX systems according to PEP 3149.  For Python 3.2 this is implemented on
    Linux, but not on OS X.

    """
    so_ext = distutils.sysconfig.get_config_var('SO') or ''
    # fix long extension for Python >=3.2, see PEP 3149.
    if not is_python_ext and 'SOABI' in distutils.sysconfig.get_config_vars():
        # Does nothing unless SOABI config var exists
        so_ext = so_ext.replace(
            '.' + distutils.sysconfig.get_config_var('SOABI'), '', 1)

    return so_ext


def load_library(libname, loader_path):
    """Load a DLL via ctypes load function. Return None on failure.

    Try loading the DLL from the current package directory first,
    then from the Windows DLL search path.

    """
    so_ext = get_shared_lib_extension()
    libname_ext = [libname + so_ext]
    if sys.version[:3] >= '3.2':
        # For Python >= 3.2 a tag may be added to lib extension
        # (platform dependent).  If we find such a tag, try both with
        # and without it.
        so_ext2 = get_shared_lib_extension(is_python_ext=True)
        if not so_ext2 == so_ext:
            libname_ext.insert(0, libname + so_ext2)

    loader_path = os.path.abspath(loader_path + "/..")

    # Need to save exception when using Python 3k, see PEP 3110.
    exc = None
    for ln in libname_ext:
        try:
            libpath = os.path.join(loader_path, ln)
            return cdll[libpath]
        except OSError as e:
            exc = e
    raise exc

    

tr = load_library('libtesserwrap', os.path.dirname(__file__))


tr.Tesserwrap_Init.restype = c_void_p
tr.Tesserwrap_Init.argtypes = [c_char_p, c_char_p]

tr.Tesserwrap_Destroy.argtypes = [c_void_p]
tr.Tesserwrap_Destroy.restype = None

tr.Tesserwrap_GetRectangle.restype = None
tr.Tesserwrap_GetRectangle.argtypes = [
    c_void_p,
    POINTER(c_ulonglong), POINTER(c_ulonglong),
    POINTER(c_ulonglong), POINTER(c_ulonglong)
]

tr.Tesserwrap_SetRectangle.restype = None
tr.Tesserwrap_SetRectangle.argtypes = [
    c_void_p,
    c_ulonglong, c_ulonglong,
    c_ulonglong, c_ulonglong
]

tr.Tesserwrap_SetImage.restype = None
tr.Tesserwrap_SetImage.argtypes = [
    c_void_p,
    c_char_p,
    c_ulonglong, c_longlong, c_longlong
]

tr.Tesserwrap_GetUTF8Text.restype = c_char_p
tr.Tesserwrap_GetUTF8Text.argtypes = [c_void_p]

tr.Tesserwrap_GetPageSegMode.restype = c_int
tr.Tesserwrap_GetPageSegMode.argtypes = [c_void_p]

tr.Tesserwrap_SetPageSegMode.restype = None
tr.Tesserwrap_SetPageSegMode.argtypes = [c_void_p, c_int]

tr.Tesserwrap_Clear.restype = None
tr.Tesserwrap_Clear.argtypes = [c_void_p]

tr.Tesserwrap_SetVariable.restype = None
tr.Tesserwrap_SetVariable.argtypes = [c_void_p, c_char_p, c_char_p]

tr.Tesserwrap_MeanTextConf.restype = c_int
tr.Tesserwrap_MeanTextConf.argtypes = [c_void_p]

class ConfidenceNode(Structure):
    pass

ConfidenceNode._fields_ = [
    ("value", c_int),
    ("next", POINTER(ConfidenceNode))
]

tr.Tesserwrap_AllWordConfidences.restype = POINTER(ConfidenceNode)
tr.Tesserwrap_AllWordConfidences.argtypes = [c_void_p]

class ResultNode(Structure):
    pass

ResultNode._fields_ = [
    ("value", c_char_p),
    ("confidence", c_float),
    ("box", c_int * 4),
    ("next", POINTER(ResultNode))
]

tr.Tesserwrap_GetResult.restype = POINTER(ResultNode)
tr.Tesserwrap_GetResult.argtypes = [c_void_p, c_int]


from .core import tr
from ctypes import c_ulonglong, byref
from collections import namedtuple
import sys
import warnings


__all__ = ["PageSegMode", "PageIteratorLevel", "Tesseract"]


class PageSegMode(object):
    PSM_OSD_ONLY = 0
    PSM_AUTO_OSD = 1
    PSM_AUTO_ONLY = 2
    PSM_AUTO = 3
    PSM_SINGLE_COLUMN = 4
    PSM_SINGLE_BLOCK_VERT_TEXT = 5
    PSM_SINGLE_BLOCK = 6
    PSM_SINGLE_LINE = 7
    PSM_SINGLE_WORD = 8
    PSM_CIRCLE_WORD = 9
    PSM_SINGLE_CHAR = 10

class PageIteratorLevel(object):
    RIL_BLOCK = 0
    RIL_PARA = 1
    RIL_TEXTLINE = 2
    RIL_WORD = 3
    RIL_SYMBOL = 4


class Tesseract(object):
    """Tesseract OCR object.

    :param datadir:
        Tesseract data-directory with Tesseract training data.

    :param lang:
        The language of the image(s) to be OCRed.

    A simple example::

        >>> from tesserwrap import Tesseract
        >>> from PIL import Image

        >>> img = Image.open("test.png")
        >>> tr = Tesseract()
        >>> tr.ocr_image(img)
        'The quick brown fox jumps ove\\n\\n'
    """

    def __init__(self, datadir="", lang="eng"):
        """Initialize a new Tesseract object

        """
        self.handle = tr.Tesserwrap_Init(
            bytes(datadir, "ascii") if sys.version[:3] >= '3.2' else datadir,
            bytes(lang, "ascii") if sys.version[:3] >= '3.2' else lang
        )

    def __del__(self):
        try:
            if self.handle and core:
                tr.Tesserwrap_Destroy(self.handle)
                self.handle = None
        except AttributeError:
            print("__del__ without handle release")
            pass

    def set_image(self, image):
        '''Takes a PIL Image and loads it into Tesseract for further
        operations.

        Note:: This function will automatically convert the image to
        Grayscale.

        :param image: image
            Image to use in tesseract.
        '''
        if image.mode != "L":
            image = image.convert("L")

        if hasattr(image, "tobytes"):
            img_bytes = image.tobytes()
        else:
            img_bytes = image.tostring()
            
        tr.Tesserwrap_SetImage(
            self.handle,
            img_bytes,                  # Image data
            len(img_bytes),             # size of buffer
            image.size[0],              # Width
            image.size[1])              # Height

    def get_text(self):
        """Get the text of the OCR'd image as a byte-string
        """
        return tr.Tesserwrap_GetUTF8Text(self.handle)

    def get_utf8_text(self):
        """Get the text of the OCR'd image as a string.

        This function is kept for backwards compatability with the 0.0
        version of tesserwrap.
        """
        return self.get_text().decode(encoding="UTF-8")

    def ocr_image(self, image):
        """OCR an image returning the UTF8 text data.

        :param image: image
            Image to be OCR'd by tesseract.
        """
        self.set_image(image)
        self.set_page_seg_mode(PageSegMode.PSM_SINGLE_BLOCK)
        return self.get_utf8_text()

    def get_rectangle(self):
        """Get the bounding rectangle that tesseract is looking at inside
        of the image.
        """
        left, top = c_ulonglong(), c_ulonglong()
        width, height = c_ulonglong(), c_ulonglong()
        tr.Tesserwrap_GetRectangle(
            self.handle,
            byref(left), byref(top),
            byref(width), byref(height))
        return ((left.value, top.value), (width.value, height.value))

    def set_rectangle(self, left, top, width, height):
        """Set the OCR detection bounding-box.

        :param left: integer
            Pixels offset right from left of the image.

        :param top: integer
            Pixels offset down from the top of the image.

        :param width: integer
            Width of the bounding-box.

        :param height: integer
            Height of the bounding-box.
        """
        tr.Tesserwrap_SetRectangle(
            self.handle,
            left, top,
            width, height)

    def get_page_seg_mode(self):
        """Returns the page analysis mode from Tesseract"""
        return tr.Tesserwrap_GetPageSegMode(self.handle)

    def set_page_seg_mode(self, mode=PageSegMode.PSM_SINGLE_BLOCK):
        """Set the page layout analysis mode.

        :param mode: integer
            The page layout analysis mode. See PageSegMode class for options
        """
        tr.Tesserwrap_SetPageSegMode(self.handle, mode)

    def clear(self):
        """Clear the tesseract Image, and clean up any Tesseract run-data."""
        tr.Tesserwrap_Clear(self.handle)

    def set_variable(self, key, value):
        """Set an internal Tesseract variable.

        :param key: str
            Variable name to change.

        :param value: str
            New variable value.

        """
        tr.Tesserwrap_SetVariable(
            self.handle,
            bytes(key, "ascii") if sys.version[:3] >= '3.2' else key,
            bytes(value, "ascii") if sys.version[:3] >= '3.2' else value
        )

    def get_mean_confidence(self):
        """Returns the (average) confidence value between 0 and 100. 
        """
        return tr.Tesserwrap_MeanTextConf(self.handle)

    def get_all_word_confidences(self):
        node = tr.Tesserwrap_AllWordConfidences(self.handle)
        result = []
        
        while bool(node):
            result.append(node.contents.value)
            node = node.contents.next

        return result

    def get_result(self, level):
        node = tr.Tesserwrap_GetResult(self.handle, level)
        result = []
        Item = namedtuple('Item', ['value', 'confidence', 'box'])

        while bool(node):
            item = Item(
                value=node.contents.value,
                confidence=node.contents.confidence,
                box = tuple(node.contents.box)
            )
            result.append(item)
            node = node.contents.next

        return result

    def get_words(self):
        """Get a list containing all the words in the OCR'd image.
        :returns: A list containing objects with the attributes:
            value: the string value of the word
            box: left, upper, right, and lower pixel coordinate
            confidence: confidence value between 0 and 100 
        """        
        return self.get_result(PageIteratorLevel.RIL_WORD)

    def get_symbols(self):
        """Get a list containing all symbols in the OCR'd image.
        :returns: A list containing objects with the attributes:
            value: the string value of the symbol
            box: left, upper, right, and lower pixel coordinate
            confidence: confidence value between 0 and 100 
        """                
        return self.get_result(PageIteratorLevel.RIL_SYMBOL)

    def get_textlines(self):
        """Get a list containing all lines in the OCR'd image.
        :returns: A list containing objects with the attributes:
            value: the string value of the line
            box: left, upper, right, and lower pixel coordinate
            confidence: confidence value between 0 and 100 
        """        
        return self.get_result(PageIteratorLevel.RIL_TEXTLINE)




def tesseract(*args, **kwargs):
    """When the lower-case version of tesseract is called, spit out a
        DeprecationWarning and create the new class object.
    """
    warnings.warn(
        "Soon 'tesseract' will be deprecated, use Tesseract instead",
        DeprecationWarning, stacklevel=2)
    return Tesseract(*args, **kwargs)



import os
import sys
from setuptools import setup, Extension

import multiprocessing
import distutils.util as du_util


# Library locator function
# Looks to see which library is available to link against
def check_lib_by_name(lib_name, search_path=None):
    s_path = ""
    platform_opts = ""
    if search_path:
        for path in search_path:
            s_path = s_path + "-L%s" % path

    # OSX specific (From: jmel - Tesserwrap: #11)
    if "macosx" in du_util.get_platform():
        platform_opts = "-arch x86_64 -execute -macosx_version_min 10.6 -pie -lm -lpthread -lcrt1.o"

    return os.system('ld %s %s -l%s' % (s_path, platform_opts, lib_name)) == 0


def find_closest_libname(lib_names, search_path=None):
    for lib_name in lib_names:
        if check_lib_by_name(lib_name, search_path):
            return lib_name
    raise Exception(
        "Cannot find Tesseract via ldconfig, confirm it is installed.")


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

extra_lib_paths = ['/usr/local/lib']

tesseract_possible_names = ['tesseract_api', 'tesseract']
tesseract_lib_name = find_closest_libname(
    tesseract_possible_names,
    extra_lib_paths)

tesser_cpp = Extension(
    'libtesserwrap',
    include_dirs=['/usr/local/include'],
    libraries=[tesseract_lib_name],
    library_dirs=extra_lib_paths,
    sources=[
        'tesserwrap/cpp/tesseract_ext.cpp',
        'tesserwrap/cpp/tesseract_wrap.cpp'],
    depends=[
        'tesserwrap/cpp/tesseract_wrap.cpp',
        'tesserwrap/cpp/tesseract_wrap.h',
        'tesserwrap/cpp/tesseract_ext.cpp',
        'tesserwrap/cpp/tesseract_ext.h'
    ],
)

if os.environ.get('READTHEDOCS', None) == 'True':
    extensions = None
else:
    extensions = [tesser_cpp]

setup(
    name="tesserwrap",
    version="0.1.5",
    author="Greg Jurman, and others",
    author_email="gdj2214@rit.edu",
    description=("Basic python bindings to the Tesseract C++ API"),
    license="Apache License 2.0",
    keywords="tesseract ocr cpp",
    url="https://github.com/gregjurman/tesserwrap",
    packages=['tesserwrap'],
    zip_safe=False,
    ext_modules=extensions,
    long_description=read('README'),
    tests_require=['nose', 'Pillow'],
    test_suite="nose.collector",
    classifiers=[
        "Topic :: Scientific/Engineering :: Image Recognition",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Libraries :: Python Modules",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
    ],
)
